from dataclasses import dataclass
from pathlib import Path

import pytest

import cordage


@dataclass
class ConfigWithoutOutputDir:
    a: int
    b: str


def test_func_with_output_dir_path(global_config):
    def func(config: ConfigWithoutOutputDir, output_dir: Path):
        assert config.a == 1
        assert config.b == "test"
        assert isinstance(output_dir, Path)
        assert output_dir.exists()

    cordage.run(func, args=["--a", "1", "--b", "test"], global_config=global_config)


def test_func_with_output_dir_str(global_config):
    def func(config: ConfigWithoutOutputDir, output_dir: str):
        assert config.a == 1
        assert config.b == "test"
        assert isinstance(output_dir, str)

    cordage.run(func, args=["--a", "1", "--b", "test"], global_config=global_config)


def test_config_with_output_dir_path(global_config):
    @dataclass
    class ConfigWithOutputDir:
        a: int
        b: str
        output_dir: Path

    def func(config: ConfigWithOutputDir):
        assert config.a == 1
        assert config.b == "test"
        assert str(config.output_dir) != "default"
        assert isinstance(config.output_dir, Path)
        assert config.output_dir.exists()

    cordage.run(func, args=["--a", "1", "--b", "test"], global_config=global_config)


def test_config_with_output_dir_str(global_config):
    @dataclass
    class ConfigWithOutputDir:
        a: int
        b: str
        output_dir: str

    def func(config: ConfigWithOutputDir):
        assert config.a == 1
        assert config.b == "test"
        assert config.output_dir != "default"
        assert isinstance(config.output_dir, str)
        assert Path(config.output_dir).exists()

    cordage.run(func, args=["--a", "1", "--b", "test"], global_config=global_config)


def test_config_and_func_with_output_dir(global_config):
    @dataclass
    class ConfigWithOutputDir:
        a: int
        b: str
        output_dir: str

    def func(config: ConfigWithOutputDir, output_dir: Path):
        assert config.a == 1
        assert config.b == "test"
        assert isinstance(config.output_dir, str)
        assert isinstance(output_dir, Path)
        assert str(output_dir) == str(config.output_dir)
        assert config.output_dir != "default"
        assert Path(config.output_dir).exists()
        assert output_dir.exists()

    cordage.run(func, args=["--a", "1", "--b", "test"], global_config=global_config)


def test_incorrect_output_dir_type(global_config):
    @dataclass
    class ConfigWithOutputDir:
        a: int
        b: str
        output_dir: int = 1

    def func(config: ConfigWithOutputDir):
        pass

    with pytest.raises(TypeError):
        cordage.run(func, args=["--a", "1", "--b", "test"], global_config=global_config)


def test_manual_output_dir(global_config):
    @dataclass
    class ConfigWithOutputDir:
        a: int
        b: str
        output_dir: str

    def func(config: ConfigWithOutputDir, output_dir: Path):
        assert isinstance(config.output_dir, str)
        assert isinstance(output_dir, Path)

        assert str(output_dir) == str(config.output_dir)
        assert output_dir.exists()

        assert output_dir == global_config.base_output_dir / "test_name"

    cordage.run(
        func,
        args=[
            "--a",
            "1",
            "--b",
            "test",
            "--output_dir",
            str(global_config.base_output_dir / "test_name"),
        ],
        global_config=global_config,
    )



from dataclasses import dataclass
from datetime import datetime, timezone
from os import PathLike
from pathlib import Path
from typing import Any, Union

from cordage.util import from_dict as config_from_dict
from cordage.util import from_file as config_from_file
from cordage.util import logger

_warned_deprecated_nested_global_config: bool = False


@dataclass
class GlobalConfig:
    """Holds the configuration for cordage."""

    PROJECT_SPECIFIC_CONFIG_PATH = Path("./cordage_configuration.json")
    GLOBAL_CONFIG_PATH: Path = Path("~/.config/cordage.json")

    base_output_dir: Path = Path("results")

    output_dir_format: str = "{start_time:%Y-%m}/{start_time:%Y-%m-%d_%H-%M-%S}{collision_suffix}"

    overwrite_existing: bool = False

    _series_spec_key = "__series__"
    _trial_indices_key = "__series-trial-indices"
    _experiment_comment_key = "__cordage-comment__"
    _output_dir_key = "__output-dir__"

    strict_mode: bool = True

    config_only: bool = False

    # Determines the logging behavior
    logging_use: bool = True
    logging_to_stream: bool = True
    logging_to_file: bool = True
    logging_filename: str = "cordage.log"

    # Determines the names of the parameters in the function to be
    # called by cordage.
    param_name_config: str = "config"
    param_name_output_dir: str = "output_dir"
    param_name_trial_object: str = "cordage_trial"

    catch_exception: bool = True

    def __post_init__(self):
        super().__init__()

        self.validate_format_strings()

    def validate_format_strings(self):
        dummy_time = datetime.now(timezone.utc).astimezone()

        # check the format strings
        self.output_dir_format.format(
            function="some_function",
            collision_suffix="_2",
            start_time=dummy_time,
        )

    @classmethod
    def resolve(cls, global_config: Union[str, PathLike, dict, "GlobalConfig", None]):
        # Dictionary: create configuration based on these values
        if isinstance(global_config, dict):
            logger.debug("Creating global from dictionary.")
            return config_from_dict(cls, global_config)

        # Path: load configuration file from this path
        elif isinstance(global_config, (str, Path)):
            global_config = Path(global_config)
            if not global_config.exists():
                msg = f"Given cordage configuration path ({global_config}) does not exist."
                raise FileNotFoundError(msg)

            logger.debug("Loading global config from file (%s).", global_config)
            return config_from_file(cls, global_config)

        # GlobalConfig object
        elif isinstance(global_config, cls):
            return global_config

        # None: look for files
        elif global_config is None:
            # Go through config file order

            # 1. Check if a project specific configuration file exists
            if cls.PROJECT_SPECIFIC_CONFIG_PATH.exists():
                logger.debug(
                    "Loading project specific global config (%s).",
                    cls.PROJECT_SPECIFIC_CONFIG_PATH,
                )
                return config_from_file(cls, cls.PROJECT_SPECIFIC_CONFIG_PATH)

            # 2. Check if a global configuration file exists
            elif cls.GLOBAL_CONFIG_PATH.exists():
                logger.debug("Loading global config (%s).", cls.GLOBAL_CONFIG_PATH)
                return config_from_file(cls, cls.GLOBAL_CONFIG_PATH)

            # 3. Use the default values
            else:
                logger.info(
                    "No cordage configuration given. Using default values. Use a project specific "
                    "(%s) or global configuration (%s) to change the behavior.",
                    cls.PROJECT_SPECIFIC_CONFIG_PATH,
                    cls.GLOBAL_CONFIG_PATH,
                )
            return cls()
        else:
            msg = "`global_config` must be one of str, PathLike, dict, cordage.GlobalConfig, None"
            raise TypeError(msg)

    @classmethod
    def _convert_old_to_new(cls, d: dict[str, Any]) -> dict[str, Any]:
        global _warned_deprecated_nested_global_config  # noqa: PLW0603

        if "logging" in d["global_config"]:
            if not _warned_deprecated_nested_global_config:
                logger.warning("Using deprecated nested global_config format.")
                _warned_deprecated_nested_global_config = True

            for k, v in d["global_config"]["logging"].items():
                d["global_config"][f"logging_{k}"] = v

            for k, v in d["global_config"]["param_names"].items():
                d["global_config"][f"param_name_{k}"] = v

            del d["global_config"]["logging"]
            del d["global_config"]["param_names"]
        return d



from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional, Union

import pytest
from config_classes import LongConfig as Config

import cordage
import cordage.exceptions


def test_simple_config(global_config):
    def func(config: Config):
        """short_function_description

        long_description

        :param config: Configuration to use.
        """
        assert config.a == 1
        assert isinstance(config.b, Path)
        assert config.c == "test"

    cordage.run(func, args=["--a", "1", "--b", "~"], global_config=global_config)


def test_loading(global_config, resources_path):
    def func(config: Config):
        assert config.a == 1
        assert isinstance(config.b, Path)
        assert config.c == "some_other_value"

    config_file = resources_path / "simple_a.toml"

    cordage.run(
        func, args=[str(config_file), "--c", "some_other_value"], global_config=global_config
    )


def test_literal_fields(global_config, resources_path):
    def func(config: Config):
        pass

    config_file = resources_path / "simple_b.json"

    with pytest.raises(cordage.exceptions.WrongTypeError):
        cordage.run(func, args=[str(config_file)], global_config=global_config)


def test_tuple_length_fields(global_config, resources_path):
    def func(config: Config):
        pass

    config_file = resources_path / "simple_c.toml"

    with pytest.raises(cordage.exceptions.WrongTypeError):
        cordage.run(func, args=[str(config_file)], global_config=global_config)


@pytest.mark.skip(reason="dacite currently does not properly support mixed tuples")
def test_valid_mixed_tuple(global_config, resources_path):
    def func(config: Config):
        pass

    config_file = resources_path / "simple_d.json"

    cordage.run(func, args=[str(config_file)], global_config=global_config)


@pytest.mark.skip(reason="dacite currently does not properly support mixed tuples")
def test_invalid_mixed_tuple(global_config, resources_path):
    def func(config: Config):
        pass

    config_file = resources_path / "simple_e.toml"

    with pytest.raises(cordage.exceptions.WrongTypeError):
        cordage.run(func, args=[str(config_file)], global_config=global_config)


def test_valid_optional(global_config, resources_path):
    def func(config: Config):
        pass

    config_file = resources_path / "simple_f.json"

    cordage.run(func, args=[str(config_file)], global_config=global_config)


def test_non_init_field(global_config):
    @dataclass
    class NonInitConfig:
        a: int
        b: float = field(init=False)

        def __post_init__(self):
            self.b = float(self.a)

    def func(config: NonInitConfig):
        assert int(config.b) == config.a

    cordage.run(func, args=["--a", "2"], global_config=global_config)

    # Invoking the command with b should make the parser exit
    with pytest.raises(SystemExit):
        cordage.run(func, args=["--a", "2", "--b", "3"], global_config=global_config)


def test_non_init_union_field(global_config):
    @dataclass
    class NonInitConfig:
        a: int
        b: Union[float, int, None] = field(init=False)

        def __post_init__(self):
            if self.a > 4:
                self.b = float(self.a)
            elif self.a > 0:
                self.b = self.a
            else:
                self.b = None

    def func(config: NonInitConfig):
        if config.a > 4:
            assert isinstance(config.b, float)
        elif config.a <= 0:
            assert config.b is None

    cordage.run(func, args=["--a", "2"], global_config=global_config)
    cordage.run(func, args=["--a", "-1"], global_config=global_config)

    # Invoking the command with b should make the parser exit
    with pytest.raises(SystemExit):
        cordage.run(func, args=["--a", "2", "--b", "3"], global_config=global_config)


def test_non_init_field_series(global_config, resources_path):
    @dataclass
    class NonInitConfig:
        a: int = -1
        b: float = field(init=False)

        def __post_init__(self):
            self.b = float(self.a)

    def func(config: NonInitConfig):
        assert int(config.b) == config.a

    cordage.run(
        func,
        args=[
            str(resources_path / "series_simple.yaml"),
        ],
        global_config=global_config,
    )


@pytest.mark.skip(
    reason="issue within dependency (see: https://github.com/konradhalas/dacite/issues/244 )"
)
def test_non_init_optional_field(global_config):
    @dataclass
    class NonInitConfig:
        a: int
        b: Optional[float] = field(init=False)

        def __post_init__(self):
            self.b = float(self.a)

    def func(config: NonInitConfig):
        assert config.b is not None
        assert int(config.b) == config.a

    cordage.run(func, args=["--a", "2"], global_config=global_config)

    # Invoking the command with b should make the parser exit
    with pytest.raises(SystemExit):
        cordage.run(func, args=["--a", "2", "--b", "3"], global_config=global_config)



import logging
import sys
from dataclasses import replace
from os import PathLike
from typing import Callable, Optional, Union

import cordage.exceptions
from cordage.context import FunctionContext
from cordage.experiment import Experiment, Metadata, Series, Status, Trial
from cordage.global_config import GlobalConfig

logger = logging.getLogger("cordage")


def run(
    func: Callable,
    args: Optional[list[str]] = None,
    *,
    description: Optional[str] = None,
    config_cls: Optional[type] = None,
    global_config: Union[PathLike, dict, GlobalConfig, None] = None,
    **kw,
) -> Experiment:
    try:
        _global_config = replace(GlobalConfig.resolve(global_config), **kw)
        context = FunctionContext(
            func,
            description=description,
            config_cls=config_cls,
            global_config=_global_config,
        )
        experiment = context.parse_args(args)
        context.execute(experiment)
        return experiment
    except cordage.exceptions.CordageError as e:
        if _global_config.catch_exception:
            logger.critical(str(e))
            sys.exit(1)
        else:
            raise


__all__ = [
    "Experiment",
    "FunctionContext",
    "GlobalConfig",
    "Metadata",
    "Series",
    "Status",
    "Trial",
    "run",
]



import logging
import shutil
from collections.abc import Generator, Iterable
from copy import deepcopy
from datetime import datetime, timezone
from itertools import chain, count, product
from json.decoder import JSONDecodeError
from math import ceil, floor, log10
from os import PathLike, getpid
from pathlib import Path
from traceback import format_exception
from typing import (
    Any,
    Generic,
    Literal,
    Optional,
    TypeVar,
    Union,
    overload,
)

from dacite import DaciteError

try:
    import colorlog
except ImportError:
    colorlog = None  # type: ignore

import typing

from cordage.metadata import Annotatable, Metadata, Status
from cordage.util import (
    TrialIndices,
    TrialIndicesEntry,
    config_output_dir_type,
    flattened_items,
    from_dict,
    logger,
    nest_items,
    nested_update,
)

if typing.TYPE_CHECKING:
    from _typeshed import DataclassInstance


ConfigClass = TypeVar("ConfigClass", bound="DataclassInstance")


class Experiment(Annotatable):
    def __init__(self, *args, config_cls: Optional[type] = None, **kwargs):
        super().__init__(*args, **kwargs)

        self.config_cls = config_cls
        self.log_handlers: list[logging.Handler] = []

    def __repr__(self):
        if self.metadata.output_dir is not None:
            return f"{self.__class__.__name__} ({self.output_dir}, status: {self.status})"
        else:
            return f"{self.__class__.__name__} (status: {self.status})"

    @property
    def log_path(self):
        return self.output_dir / self.global_config.logging_filename

    @property
    def status(self) -> Status:
        return self.metadata.status

    def has_status(self, *status: Status):
        return len(status) == 0 or self.status in status

    @property
    def result(self) -> Any:
        return self.metadata.result

    def start(self):
        """Start the execution of an experiment.

        Set start time, create output directory, registers run, etc.
        """
        assert self.config_cls is not None
        assert not self.status.has_started, f"{self.__class__.__name__} has already been started."
        self.metadata.start_time = datetime.now(timezone.utc).astimezone()
        self.metadata.status = Status.RUNNING
        self.metadata.additional_info["process_id"] = getpid()
        self.create_output_dir()
        self.save_metadata()
        self.save_annotations()
        self.setup_log()

    def end(self, status: Status):
        """End the execution of an experiment.

        Write metadata, close logs, etc.
        """
        self.metadata.end_time = datetime.now(timezone.utc).astimezone()
        self.metadata.status = status
        self.save_metadata()
        self.save_annotations()
        self.teardown_log()

    def handle_exception(self, exc_type, exc_value, traceback):
        traceback_string = "".join(format_exception(exc_type, value=exc_value, tb=traceback))

        logger.exception("", exc_info=(exc_type, exc_value, traceback))
        self.metadata.additional_info["exception"] = {
            "short": repr(exc_value),
            "traceback": traceback_string,
        }

    def __enter__(self):
        self.start()
        logger.info("%s '%s' started.", self.__class__.__name__, str(self.output_dir))

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            logger.info("%s '%s' completed.", self.__class__.__name__, str(self.output_dir))
            self.end(status=Status.COMPLETE)
        elif issubclass(exc_type, KeyboardInterrupt):
            logger.warning("%s '%s' aborted.", self.__class__.__name__, str(self.output_dir))
            self.end(status=Status.ABORTED)
            return False
        else:
            self.handle_exception(exc_type, exc_value, traceback)
            logger.warning("%s '%s' failed.", self.__class__.__name__, str(self.output_dir))
            self.end(status=Status.FAILED)
            return False

    def load_data(self):
        """Synchronize to existing output directory."""
        assert self.metadata.output_dir is not None, (
            f"Cannot synchronize a {self.__class__.__name__} which has no `output_dir`."
        )

        if self.metadata.output_dir.exists():
            metadata = self.load_metadata(self.metadata.output_dir)

            if not isinstance(self.metadata.configuration, dict):
                metadata.configuration = from_dict(
                    type(self.metadata.configuration), metadata.configuration
                )

            self.metadata = metadata

            self.load_annotations()
        else:
            logger.warning(
                "Experiment directory (%s) not found. Cannot load the data.",
                str(self.metadata.output_dir),
            )

        return self

    @classmethod
    def from_path(
        cls,
        path: PathLike,
        *,
        config_cls: Optional[type[ConfigClass]] = None,
        load_series_trials: bool = True,
    ):
        metadata: Metadata = cls.load_metadata(path)

        experiment: Experiment
        if not metadata.is_series:
            experiment = Trial(metadata, config_cls=config_cls)

        else:
            experiment = Series(metadata, config_cls=config_cls)
            if load_series_trials:
                for trial in experiment:
                    if trial.metadata.output_dir.exists():
                        trial.load_data()

        experiment.load_annotations()

        return experiment

    @classmethod
    def all_from_path(
        cls, results_path: Union[str, PathLike], *, skip_hidden: bool = True
    ) -> list["Experiment"]:
        """Load all experiments from the results_path."""
        results_path = Path(results_path)

        seen_dirs: set[Path] = set()
        experiments = []

        for p in results_path.rglob("*/cordage.json"):
            if skip_hidden and any(
                part.startswith(".") for part in (p.relative_to(results_path)).parts
            ):
                continue

            path = p.parent

            if path.parent in seen_dirs:
                # we already encountered a parent experiment (series)
                continue

            seen_dirs.add(path)

            try:
                experiments.append(cls.from_path(p.parent))
            except (JSONDecodeError, DaciteError) as exc:
                logger.warning("Couldn't load '%s': %s", str(path), str(exc))

        return sorted(experiments, key=lambda exp: exp.output_dir)

    def setup_log(self):
        logger = logging.getLogger()

        if not self.global_config.logging_use:
            return

        # in this case, a StreamHandler was set up by the series
        is_toplevel = self.metadata.parent_dir is None

        handler: logging.Handler

        if self.global_config.logging_to_stream and is_toplevel:
            # add colored stream handler
            format_str = "%(name)s:%(filename)s:%(lineno)d - %(message)s"

            if colorlog is not None:
                handler = colorlog.StreamHandler()
                handler.setFormatter(
                    colorlog.ColoredFormatter(
                        f"%(log_color)s%(levelname)-8s%(reset)s {format_str}"
                    )
                )
            else:
                handler = logging.StreamHandler()
                handler.setFormatter(logging.Formatter(f"%(levelname)-8s {format_str}"))

            logger.addHandler(handler)
            self.log_handlers.append(handler)

        if self.global_config.logging_to_file:
            # setup logging to local output_dir
            formatter = logging.Formatter(
                "%(asctime)s %(levelname)-8s %(name)s:%(filename)s:%(lineno)d - %(message)s"
            )
            handler = logging.FileHandler(self.log_path)
            handler.setFormatter(formatter)

            logger.addHandler(handler)
            self.log_handlers.append(handler)

    def teardown_log(self):
        logger = logging.getLogger()

        for handler in self.log_handlers:
            handler.close()
            logger.removeHandler(handler)

    def create_output_dir(self):
        if self.metadata.output_dir is not None:
            self.output_dir.mkdir(parents=True, exist_ok=True)
            self.set_output_dir(self.output_dir)
            return self.output_dir

        tried_paths: set[Path] = set()
        suffix = ""

        for i in count(1):
            if i > 1:
                level = floor(log10(i) / 2) + 1
                suffix = "_" * level + str(i).zfill(2 * level)

            path = (
                self.global_config.base_output_dir
                / self.global_config.output_dir_format.format(
                    **self.metadata.__dict__,
                    collision_suffix=suffix,
                )
            )

            if path in tried_paths:
                # suffix was already tried: assume that further tries
                # wont resolve this collision
                msg = f"Path {path} does already exist - collision could not be avoided."
                raise RuntimeError(msg)

            try:
                path.mkdir(parents=True, exist_ok=False)
                self.set_output_dir(path)
                return path
            except FileExistsError:
                if self.global_config.overwrite_existing:
                    logger.warning(
                        "Path %s does existing. Replacing directory with new one.", str(path)
                    )
                    shutil.rmtree(path)
                    path.mkdir(parents=True)
                    self.set_output_dir(path)
                    return path
                else:
                    tried_paths.add(path)


class Trial(Experiment, Generic[ConfigClass]):
    def __init__(
        self,
        metadata: Optional[Metadata] = None,
        /,
        config: Optional[dict[str, Any]] = None,
        config_cls=None,
        **kw,
    ):
        if metadata is not None:
            if len(kw) == 0 and config is None:
                super().__init__(metadata, config_cls=config_cls)
            else:
                msg = "If metadata are provided, config and additional keywords can not be set."
                raise TypeError(msg)
        else:
            super().__init__(configuration=config, config_cls=config_cls, **kw)

        self._config: Optional[ConfigClass] = None

    @property
    def config(self) -> ConfigClass:
        if self._config is None:
            if self.config_cls is None:
                msg = (
                    "`trial.config` is only available if the configuration was loaded with a "
                    "configuration dataclass. You could use `trial.metadata.configuration` "
                    "instead or pass `config_cls` to the trial initializer."
                )
                raise AttributeError(msg)

            if self.metadata.output_dir is not None:
                self.set_output_dir(self.metadata.output_dir)

            # Create the config object
            self._config = from_dict(
                self.config_cls,
                self.metadata.configuration,
                strict=self.metadata.global_config.strict_mode,
            )

        return self._config

    def set_output_dir(self, path: Path):
        super().set_output_dir(path)

        if self.config_cls is not None:
            output_dir_type = config_output_dir_type(
                self.config_cls, self.global_config.param_name_output_dir
            )

            if output_dir_type is not None:
                self.metadata.configuration["output_dir"] = path

                # Config has attribute output_dir, mypy does not know it
                if self._config is not None:
                    self.config.output_dir = output_dir_type(path)  # type: ignore


class Series(Generic[ConfigClass], Experiment):
    trials: list[Trial[ConfigClass]]

    def __init__(
        self,
        metadata: Optional[Metadata] = None,
        /,
        base_config: Optional[dict[str, Any]] = None,
        series_spec: Union[list[dict], dict[str, list], None] = None,
        trial_indices: Optional[TrialIndices] = None,
        config_cls=None,
        **kw,
    ):
        if metadata is not None:
            assert len(kw) == 0 and base_config is None and series_spec is None
            super().__init__(metadata, config_cls=config_cls)
        else:
            if isinstance(series_spec, list):
                series_spec = [
                    nest_items(flattened_items(trial_update, sep="."))
                    for trial_update in series_spec
                ]

            super().__init__(
                configuration={
                    "base_config": base_config,
                    "series_spec": series_spec,
                    "trial_indices": trial_indices,
                },
                config_cls=config_cls,
                **kw,
            )

        self.validate_series_spec()
        self.make_all_trials()

    def validate_series_spec(self):
        assert (
            self.metadata.configuration.get("series_skip", None) is None
            or self.metadata.configuration.get("series_trial", None) is None
        ), "Only one of `series_skip` and`series_trial` should be used."

        series_spec = self.series_spec

        if isinstance(series_spec, list):
            for config_update in series_spec:
                assert isinstance(config_update, dict)

        elif isinstance(series_spec, dict):

            def only_list_nodes(d):
                for v in d.values():
                    if isinstance(v, dict):
                        if not only_list_nodes(v):
                            return False
                    elif not isinstance(v, list):
                        return False
                    return True

            assert only_list_nodes(series_spec), f"Invalid series specification: {series_spec}"
        else:
            assert series_spec is None

    @property
    def base_config(self) -> dict[str, Any]:
        return self.metadata.configuration["base_config"]

    @property
    def series_spec(self) -> Union[list[dict], dict[str, list], None]:
        return self.metadata.configuration["series_spec"]

    @property
    def series_skip(self) -> int:
        skip: Optional[int] = self.metadata.configuration.get("series_skip", None)

        if skip is None:
            return 0
        else:
            return skip

    @property
    def is_singular(self):
        return self.series_spec is None

    def __enter__(self):
        for i, trial in enumerate(self.trials, start=1):
            if i <= self.series_skip:
                trial.metadata.status = Status.SKIPPED
            else:
                trial.metadata.status = Status.PENDING

        if not self.is_singular:
            super().__enter__()
        # else: do nothing

    def __exit__(self, *args):
        if not self.is_singular:
            super().__exit__(*args)
        # else: do nothing

    @overload
    def get_changing_fields(self, sep: Literal[None] = None) -> set[tuple[Any, ...]]: ...

    @overload
    def get_changing_fields(self, sep: str) -> set[str]: ...

    def get_changing_fields(
        self, sep: Optional[str] = None
    ) -> Union[set[tuple[Any, ...]], set[str]]:
        keys: set = set()

        if isinstance(self.series_spec, list):
            for trial_update in self.series_spec:
                for k, _ in flattened_items(trial_update, sep=sep):
                    keys.add(k)

        elif isinstance(self.series_spec, dict):
            for k, _ in flattened_items(self.series_spec, sep=sep):
                keys.add(k)

        return keys

    def get_trial_updates(self) -> Generator[dict, None, None]:
        if isinstance(self.series_spec, list):
            yield from self.series_spec
        elif isinstance(self.series_spec, dict):
            keys, values = zip(*flattened_items(self.series_spec, sep="."))

            for update_values in product(*values):
                yield nest_items(zip(keys, update_values))
        else:
            yield {}

    def _derive_len(self) -> int:
        if isinstance(self.series_spec, list):
            return len(self.series_spec)
        elif isinstance(self.series_spec, dict):
            num_trials = 1
            for _, values in flattened_items(self.series_spec):
                num_trials *= len(values)
            return num_trials
        else:
            return 1

    def __len__(self) -> int:
        if len(self.trials) != self._derive_len():
            msg = (
                f"Number of existing ({len(self.trials)}) and expected trials "
                f"({self._derive_len()}) do not match."
            )
            raise RuntimeError(msg)

        return len(self.trials)

    def make_trial(self, **kw):
        additional_info = kw.pop("additional_info", None)

        fields_to_update: dict[str, Any] = {
            "output_dir": None,
            "configuration": {},
            "additional_info": {},
            "status": Status.UNKOWN,
            "parent_dir": None,
            **kw,
        }

        trial_metadata = self.metadata.replace(**fields_to_update)

        if additional_info is not None:
            assert isinstance(additional_info, dict)
            trial_metadata.additional_info.update(additional_info)

        return Trial(trial_metadata, config_cls=self.config_cls)

    def make_all_trials(self):
        if self.series_spec is None:
            # single trial experiment
            logger.debug("Configuration yields a single experiment.")
            single_trial = self.make_trial(configuration=self.base_config)
            single_trial.annotations = self.annotations

            if self.metadata.output_dir is not None:
                single_trial.set_output_dir(self.metadata.output_dir)

            self.trials = [single_trial]

        else:
            logger.debug(
                "The given configuration yields an experiment series with %d experiments.",
                self._derive_len(),
            )
            self.trials = []

            for i, trial_update in enumerate(self.get_trial_updates(), start=1):
                trial_configuration: dict[str, Any] = deepcopy(self.base_config)

                nested_update(trial_configuration, trial_update)

                trial = self.make_trial(
                    configuration=trial_configuration,
                    additional_info={"trial_index": i},
                )
                self.trials.append(trial)

    def __iter__(self):
        return self.get_all_trials(include_skipped=False)

    def get_effective_indices(self, *, include_skipped=False) -> list[int]:
        if include_skipped:
            return list(range(1, len(self) + 1))

        entries: Optional[TrialIndices] = self.metadata.configuration.get("trial_indices", None)

        if not entries:
            return list(range(1, len(self) + 1))

        indices: list[Union[Iterable[int]]] = []

        entry: TrialIndicesEntry
        for entry in entries:
            if isinstance(entry, int):
                indices.append([entry])
            else:
                start, end = entry

                if start is None:
                    # Indices start at 1
                    start = 1

                if end is not None:
                    # Ranges are inclusive
                    end += 1

                indices.append(range(*slice(start, end).indices(len(self) + 1)))

        # Deduplicate (maintaining the insertion order)
        return list(dict.fromkeys(chain.from_iterable(indices)))

    def get_all_trials(
        self, *, include_skipped: bool = False
    ) -> Generator[Trial[ConfigClass], None, None]:
        assert self.trials is not None

        if not self.is_singular:
            for i in self.get_effective_indices(include_skipped=include_skipped):
                trial = self.trials[i - 1]
                trial_subdir = str(i).zfill(ceil(log10(len(self))))
                trial.metadata.output_dir = self.output_dir / trial_subdir
                yield trial

        else:
            assert len(self.trials) == 1
            yield self.trials[0]



from dataclasses import dataclass, field
from pathlib import Path
from typing import Literal, Optional


@dataclass
class SimpleConfig:
    a: int = 42
    b: str = "test"


@dataclass
class AlphaConfig:
    a: int
    b: str = "b_value"


@dataclass
class BetaConfig:
    a: str
    b: int = 0


@dataclass
class NestedConfig:
    """config_description.

    :param a: a_help_str
    :param d: wrong_help_text
    """

    alpha: AlphaConfig
    beta: BetaConfig = field(default_factory=lambda: BetaConfig(a="a_value"))

    a: str = "e_default"

    # these fields are used in test_more_trial_series for checking the
    # configuration and output dir etc.
    alphas: int = 1
    betas: int = 1


@dataclass
class LongConfig:
    """config_description.

    :param a: a_help_str
    :param d: wrong_help_text
    """

    a: int
    b: Path
    c: str = "test"
    d: int = field(default=1, metadata={"help": "correct_help_text"})
    e: Literal["a", "b", "c"] = "a"
    f: bool = False
    g: Optional[int] = None
    h: tuple[str, ...] = ("a", "b")
    i: tuple[str, str] = ("a", "b")
    j: tuple[str, int, float] = ("a", 1, 1.0)
    k: Optional[int] = None



import dataclasses
import json
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from _typeshed import DataclassInstance

from cordage import GlobalConfig
from cordage.util import to_dict


def test_global_config(global_config: "DataclassInstance"):
    assert dataclasses.is_dataclass(global_config)

    json.dumps(to_dict(global_config))


def test_format_string_validation(global_config):
    # This should be valid
    _ = GlobalConfig()

    with pytest.raises(KeyError):
        _ = dataclasses.replace(global_config, output_dir_format="{non_existing_field}")

    _ = dataclasses.replace(global_config, output_dir_format="{collision_suffix}")

    with pytest.raises(ValueError):
        _ = dataclasses.replace(global_config, output_dir_format="{function:%Y}")



import math

import pytest
from config_classes import NestedConfig as Config

import cordage
from cordage import Series
from cordage.context import TrialIndexMixin


def test_trial_series_list(global_config, resources_path):
    trial_store: list[cordage.Trial] = []

    def func(config: Config, cordage_trial: cordage.Trial, trial_store=trial_store):  # noqa: ARG001
        trial_store.append(cordage_trial)

    config_file = resources_path / "series_list.yml"

    series = cordage.run(func, args=[str(config_file)], global_config=global_config)

    assert isinstance(series, Series)

    assert series.get_changing_fields() == {("alpha", "a"), ("alpha", "b"), ("beta", "a")}

    for i, t in zip(range(1, 6), trial_store):
        assert t.metadata.additional_info["trial_index"] == i
        assert t.config.alpha.a == i
        assert t.config.alpha.b == f"b{i}"
        assert t.config.beta.a == f"c{i}"

        assert t.output_dir == global_config.base_output_dir / "experiment" / str(i)


@pytest.mark.parametrize("letter", "abc")
def test_more_trial_series(global_config, resources_path, letter):
    trial_store: list[cordage.Trial] = []

    def func(config: Config, cordage_trial: cordage.Trial, trial_store=trial_store):  # noqa: ARG001
        trial_store.append(cordage_trial)

    config_file = resources_path / f"series_{letter}.toml"

    cordage.run(
        func, args=[str(config_file), "--alpha.b", "b_incorrect"], global_config=global_config
    )

    assert len(trial_store) == trial_store[0].config.alphas * trial_store[0].config.betas

    for i, trial in enumerate(trial_store, start=1):
        assert trial.config.alpha.b == "b1"
        assert trial.config.beta.a == "c" + str(math.ceil(i / trial_store[0].config.alphas))
        assert trial.config.alpha.a == 1 + ((i - 1) % trial_store[0].config.alphas)

        assert trial.metadata.parent_dir is not None
        assert trial.metadata.parent_dir.parts[-1] == "experiment"

        if len(trial_store) <= 10:
            assert trial.output_dir == global_config.base_output_dir / "experiment" / f"{i}"

        else:
            assert trial.output_dir == global_config.base_output_dir / "experiment" / f"{i:02}"


def test_invalid_trial_series(global_config, resources_path):
    def func(config: Config, cordage_trial: cordage.Trial):
        pass

    config_file = resources_path / "series_invalid.json"

    with pytest.raises(ValueError):
        cordage.run(func, args=[str(config_file)], global_config=global_config)


@pytest.mark.parametrize(
    "args, expected_trials",
    [
        (("--trial-index", "4-"), (4, 5)),
        (("--trial-index", "2"), (2,)),
        (("--trial-index", "2, 3"), (2, 3)),
        (("--trial-index", "1, 3-5"), (1, 3, 4, 5)),
        (("--trial-index", "-3"), (1, 2, 3)),
    ],
)
def test_partial_series_execution(global_config, resources_path, expected_trials, args):
    trial_store: list[cordage.Trial] = []

    def func(config: Config, cordage_trial: cordage.Trial, trial_store=trial_store):  # noqa: ARG001
        trial_store.append(cordage_trial)

    config_file = resources_path / "series_list.yml"

    cordage.run(func, args=[str(config_file), *args], global_config=global_config)

    assert len(trial_store) == len(expected_trials)

    for i, t in zip(expected_trials, trial_store):
        assert t.metadata.additional_info["trial_index"] == i
        assert t.config.alpha.a == i
        assert t.config.alpha.b == f"b{i}"
        assert t.config.beta.a == f"c{i}"

        assert t.output_dir == global_config.base_output_dir / "experiment" / str(i)


@pytest.mark.parametrize(
    "text,match",
    [
        ("1-5", (1, 5)),
        (" 1-5", (1, 5)),
        ("1 -5", (1, 5)),
        ("1- 5 ", (1, 5)),
        ("7", 7),
        (" 6", 6),
        ("7 ", 7),
        (" 4 ", 4),
        ("7 ", 7),
        (" 6", 6),
        ("2-78", (2, 78)),
        ("5-", (5, None)),
        ("-8", (None, 8)),
        ("10-20", (10, 20)),
        ("-", False),
        ("-5-8", False),
        ("a-b", False),
        ("3-4-5", False),
    ],
)
def test_trial_range_entry_matching(text, match):
    m = TrialIndexMixin()
    if match is False:
        with pytest.raises(ValueError):
            m.match_trial_range_entry(text)

    else:
        assert m.match_trial_range_entry(text) == match


@pytest.mark.parametrize(
    "text,match",
    [
        ("1-5, 2-78, -8,7", [(1, 5), (2, 78), (None, 8), 7]),
        ("7, 6, 3", [7, 6, 3]),
        ("10-20, 1, 2, 5-6", [(10, 20), 1, 2, (5, 6)]),
        ("a, a-b, 4", False),
        ("1, 3-4-5, 5, 3", False),
    ],
)
def test_trial_range_matching(text, match):
    m = TrialIndexMixin()
    if not match:
        with pytest.raises(ValueError):
            m.match_trial_range(text)

    else:
        assert m.match_trial_range(text) == match



import json
import re
from collections.abc import Iterable, Mapping
from dataclasses import dataclass, field
from dataclasses import replace as dataclass_replace
from datetime import datetime
from os import PathLike
from pathlib import Path
from typing import (
    Any,
    Optional,
    TypeVar,
)

try:
    import colorlog
except ImportError:
    colorlog = None  # type: ignore

import typing
from enum import Enum

from cordage.global_config import GlobalConfig
from cordage.util import (
    from_dict,
    logger,
    to_dict,
)

if typing.TYPE_CHECKING:
    from _typeshed import DataclassInstance


ConfigClass = TypeVar("ConfigClass", bound="DataclassInstance")


class Status(str, Enum):
    UNKOWN = "unkown"
    PENDING = "pending"
    RUNNING = "running"
    COMPLETE = "complete"
    FAILED = "failed"
    ABORTED = "aborted"
    SKIPPED = "skipped"

    def __str__(self) -> str:
        return self.value

    @property
    def has_started(self):
        """The pname property."""
        return self not in (self.UNKOWN, self.PENDING)


@dataclass
class Metadata:
    function: str

    global_config: GlobalConfig

    configuration: dict[str, Any]

    output_dir: Optional[Path] = None
    status: Status = Status.UNKOWN

    start_time: Optional[datetime] = None
    end_time: Optional[datetime] = None

    result: Any = None

    parent_dir: Optional[Path] = None

    additional_info: dict = field(default_factory=dict)

    @property
    def duration(self):
        assert self.end_time is not None and self.start_time is not None

        return self.end_time - self.start_time

    def replace(self, **changes):
        return dataclass_replace(self, **changes)

    @property
    def is_series(self):
        return isinstance(self.configuration, dict) and "series_spec" in self.configuration

    def to_dict(self):
        return to_dict(self)

    @classmethod
    def from_dict(cls, data: Mapping):
        return from_dict(cls, data)


class MetadataStore:
    def __init__(
        self,
        metadata: Optional[Metadata] = None,
        /,
        global_config: Optional[GlobalConfig] = None,
        **kw,
    ):
        self.metadata: Metadata

        if metadata is not None:
            if global_config is not None or len(kw) > 0:
                msg = "Using the `metadata` argument is incompatible with using other arguments."
                raise TypeError(msg)
            else:
                self.metadata = metadata
        else:
            if global_config is None:
                global_config = GlobalConfig()

            self.metadata = Metadata(global_config=global_config, **kw)

    @property
    def global_config(self) -> GlobalConfig:
        return self.metadata.global_config

    @property
    def output_dir(self) -> Path:
        if self.metadata.output_dir is None:
            msg = f"{self.__class__.__name__} has not been started yet."
            raise RuntimeError(msg)
        else:
            return self.metadata.output_dir

    @property
    def parent_dir(self) -> Optional[Path]:
        return self.metadata.parent_dir

    def set_output_dir(self, path: Path):
        self.metadata.output_dir = path

    @property
    def metadata_path(self):
        return self.output_dir / "cordage.json"

    def save_metadata(self):
        md_dict = self.metadata.to_dict()

        with open(self.metadata_path, "w", encoding="utf-8") as fp:

            def invalid_obj_default(obj):
                logger.warning("Cannot serialize %s", str(obj))

            json.dump(md_dict, fp, indent=4, default=invalid_obj_default)

    @classmethod
    def load_metadata(cls, path: PathLike) -> Metadata:
        path = Path(path)
        if not path.suffix == ".json":
            path = path / "cordage.json"

        with path.open("r", encoding="utf-8") as fp:
            metadata_dict = GlobalConfig._convert_old_to_new(json.load(fp))
            metadata = Metadata.from_dict(metadata_dict)

        if metadata.output_dir != path.parent:
            logger.info(
                f"Output dir is not correct anymore. Changing it to the actual directory"
                f"({metadata.output_dir} -> {path.parent})"
            )
            metadata.output_dir = path.parent

        return metadata


class Annotatable(MetadataStore):
    TAG_PATTERN = re.compile(r"\B#(\w*[a-zA-Z]+\w*)")

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)

        self.annotations = {}

    @property
    def tags(self):
        tags = set(self.explicit_tags)

        # implicit tags
        tags.update(re.findall(self.TAG_PATTERN, self.comment))

        return list(tags)

    @property
    def explicit_tags(self):
        if "tags" not in self.annotations:
            self.annotations["tags"] = []
        return self.annotations["tags"]

    def add_tag(self, *tags: Iterable):
        for t in tags:
            if t not in self.explicit_tags:
                self.explicit_tags.append(t)

    def has_tag(self, *tags: str):
        return len(tags) == 0 or any(t in tags for t in self.tags)

    @property
    def comment(self):
        return self.annotations.get("comment", "") or ""

    @comment.setter
    def comment(self, value):
        self.annotations["comment"] = value

    @property
    def annotations_path(self):
        return self.output_dir / "annotations.json"

    def save_annotations(self):
        with open(self.annotations_path, "w", encoding="utf-8") as fp:
            json.dump(self.annotations, fp, indent=4)

    def load_annotations(self):
        if self.annotations_path.exists():
            with self.annotations_path.open("r", encoding="utf-8") as fp:
                self.annotations = json.load(fp)



from dataclasses import dataclass
from pathlib import Path

import pytest
from config_classes import NestedConfig
from config_classes import SimpleConfig as Config

import cordage
from cordage.util import from_file, get_nested_field, set_nested_field, to_file


@pytest.mark.parametrize("extension", ["toml", "yaml", "yml", "yl", "json"])
def test_different_extensions(tmp_path, extension):
    config_data = {"a": 10, "b": "20"}
    config = Config(a=10, b="20")

    path = tmp_path / f"data.{extension}"

    to_file(config_data, path)
    assert from_file(Config, path) == config

    to_file(config, path)
    assert from_file(Config, path) == config


def test_unkown_extensions(tmp_path):
    config = Config(a=10, b="20")

    path = tmp_path / "data.unkown"

    with pytest.raises(RuntimeError):
        to_file(config, path)

    with pytest.raises(RuntimeError):
        from_file(Config, path)


def test_value_casting(tmp_path):
    @dataclass
    class ComplexConfig:
        data: dict

    config = ComplexConfig({"p": Path("."), "i": 42, "pi": 3.14})

    path = tmp_path / "data.json"

    to_file(config, path)

    loaded = from_file(ComplexConfig, path)

    assert loaded.data["p"] == "."
    assert loaded.data["i"] == 42
    assert loaded.data["pi"] == 3.14


def test_nested_value_retrieval(global_config):
    def func(config: NestedConfig):
        assert get_nested_field(config, "alpha.a") == 42

        set_nested_field(config, "alpha.a", 123)

        assert get_nested_field(config, "alpha.a") == 123

    cordage.run(func, ["--alpha.a", "42", "--beta.a", "c_value"], global_config=global_config)



import logging

import pytest
from config_classes import LongConfig, SimpleConfig

import cordage

log = logging.getLogger(__name__)


def test_manual_output_dir(global_config, tmp_path):
    def func(config: SimpleConfig):
        pass

    experiment = cordage.run(
        func,
        args=["--output_dir", str(tmp_path / "some_specific_output_dir")],
        global_config=global_config,
    )

    assert experiment.output_dir == tmp_path / "some_specific_output_dir"


def test_manual_output_dir_for_series(global_config, tmp_path, resources_path):
    def func(config: SimpleConfig, cordage_trial, output_dir):  # noqa: ARG001
        assert "trial_index" in cordage_trial.metadata.additional_info
        assert output_dir == tmp_path / "some_specific_output_dir" / str(
            cordage_trial.metadata.additional_info["trial_index"]
        )

    cordage.run(
        func,
        args=[
            "--output_dir",
            str(tmp_path / "some_specific_output_dir"),
            str(resources_path / "series_simple.yaml"),
        ],
        global_config=global_config,
    )


def test_help(capfd, global_config):
    def func(config: LongConfig):  # noqa: ARG001
        """short_function_description

        long_description

        :param config: Configuration to use.
        """
        msg = "This should not be executed."
        raise RuntimeError(msg)

    with pytest.raises(SystemExit):
        cordage.run(func, args=["--help"], global_config=global_config)

    out, _ = capfd.readouterr()

    log.info("Captured log:\n%s", out)

    assert "a_help_str" in out
    assert "correct_help_text" in out
    assert "wrong_help_text" not in out
    assert ":param" not in out

    assert "short_function_description" in out
    assert "long_description" not in out
    assert "config_description" not in out

    first_line = out.split("\n")[0]

    assert "[config_file]" in first_line
    assert "<configuration options to overwrite>" in first_line


def test_explicit_description(capfd, global_config):
    def func(config: LongConfig):  # noqa: ARG001
        msg = "This should not be executed."
        raise RuntimeError(msg)

    with pytest.raises(SystemExit):
        cordage.run(
            func, args=["--help"], global_config=global_config, description="EXPLICIT_DESCRIPTION"
        )

    out, _ = capfd.readouterr()

    assert "EXPLICIT_DESCRIPTION" in out



import argparse
import dataclasses
import inspect
import re
import sys
from collections.abc import Mapping
from contextlib import contextmanager
from pathlib import Path
from typing import (
    Any,
    Callable,
    ClassVar,
    Literal,
    Optional,
    Union,
    get_args,
    get_origin,
)

from docstring_parser import parse as parse_docstring

from cordage.experiment import Experiment, Series, Status, Trial
from cordage.global_config import GlobalConfig
from cordage.util import (
    ConfigClass,
    TrialIndices,
    logger,
    nest_items,
    nested_update,
    read_dict_from_file,
)


class MissingType:
    def __repr__(self):
        return "<MISSING>"


MISSING = MissingType()


SUPPORTED_PRIMITIVES = (int, bool, str, float, Path)


class Singleton(type):
    _instances: ClassVar[dict[type, Any]] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class ExperimentStack(metaclass=Singleton):
    """Represents the stack of currently running experiments.

    This class is used internally, to determine whether an experiment
    was started from within another.
    """

    def __init__(self):
        self.running: list[Experiment] = []

    def push(self, experiment: Experiment):
        """Push a new experiment on the stack."""
        self.running.append(experiment)

    def pop(self) -> Experiment:
        """Pop the currently running experiment from the stack."""
        return self.running.pop()

    def peek(self) -> Optional[Experiment]:
        """Get the currently active experiment from the stack.

        If the stack is empty, None is returned.
        """
        if len(self.running) > 0:
            return self.running[-1]
        else:
            return None

    def peek_dir(self) -> Optional[Path]:
        """Get the output_dir of the currently running experiment.

        If the stack is empty, None is returned.
        """
        if len(self.running) > 0:
            return self.running[-1].output_dir
        else:
            return None

    def __len__(self):
        return len(self.running)

    @contextmanager
    def with_experiment_on_stack(self, experiment: Experiment):
        """Put a new experiment on the stack and set its `parent_dir`.

        The `parent_dir` will then point to the experiment which was
        running so far (i.e. the outer experiment)."""
        if self.peek() == experiment:
            yield experiment

        else:
            experiment.metadata.parent_dir = self.peek_dir()
            self.push(experiment)
            try:
                with experiment:
                    yield experiment
            finally:
                self.pop()


experiment_stack: ExperimentStack = ExperimentStack()


class TrialIndexMixin:
    trial_index_pattern = re.compile(
        r"^(?!-\s*$)"  # Not just a dash
        r"(?P<start>\d+)?"  # Optional range start
        r"\s*(?P<is_range>-)?"  # optional range indicator
        r"\s*(?P<end>\d+)??$"  # Optional range end (if only one number, start is used)
    )

    def match_trial_range_entry(
        self,
        value,
    ) -> Union[int, tuple[Optional[int], Optional[int]]]:
        match = self.trial_index_pattern.match(value.strip())
        if not match:
            msg = (
                f"'{value}' does not match the single index or range pattern "
                "(either 'I' or 'I-J', with positive ints I&J)."
            )
            raise ValueError(msg)

        # If '-' is included, a range is specified
        elif match.group("is_range"):
            start = match.group("start")
            if start is not None:
                start = int(start)

            end = match.group("end")
            if end is not None:
                end = int(end)

            return (start, end)

        else:
            return int(match.group("start"))

    def match_trial_range(self, value) -> TrialIndices:
        return [self.match_trial_range_entry(v) for v in value.split(",")]


class FunctionContext(TrialIndexMixin):
    """Wrapper for a function which accepts a dataclass configuration.

    This class can be used to:
    - parse argruments matching the config dataclass,
    - build a dictionary of the arguments expected by the function, and
    - call the function for each trial in a series.
    """

    global_config: GlobalConfig

    usage_str: str = "%(prog)s [-h] [config_file] <configuration options to overwrite>"

    def __init__(
        self,
        func: Callable,  # expects a dataclass
        global_config: GlobalConfig,
        description: Optional[str] = None,
        config_cls: Optional[type[ConfigClass]] = None,
    ):
        self.global_config = global_config
        self.set_function(func)
        self.set_config_cls(config_cls)
        self.set_description(description)
        self.construct_argument_parser()

    def set_description(self, description: Optional[str] = None):
        if description is None:
            if self.func.__doc__ is not None:
                self.description = parse_docstring(self.func.__doc__).short_description
            else:
                self.description = self.func_name

        else:
            self.description = description

    def set_config_cls(self, config_cls: Optional[type] = None):
        # derive configuration class
        if config_cls is None:
            self.main_config_cls = self.func_parameters[
                self.global_config.param_name_config
            ].annotation

        else:
            self.main_config_cls = config_cls

        if not dataclasses.is_dataclass(self.main_config_cls):
            msg = (
                "Configuration class could not be derived: Either pass a configuration dataclass "
                "via `config_cls` or annotate the configuration parameter "
                f"`{self.global_config.param_name_config}` with a dataclass."
            )
            raise TypeError(msg)

    @property
    def func(self) -> Callable:
        return self._func

    @property
    def func_parameters(self):
        return self._func_parameters

    @property
    def func_name(self):
        return self._func_name

    def set_function(self, func: Callable):
        self._func = func
        self._func_parameters = inspect.signature(func).parameters
        self._func_name = self.func.__name__

        if self.global_config.param_name_config not in self.func_parameters:
            msg = (
                "Callable must accept config argument (as "
                f"`{self.global_config.param_name_config}`)."
            )
            raise TypeError(msg)

    def construct_argument_parser(self):
        """Construct an argparser for a given config class."""
        # add parser arguments from dataclass

        self.argument_parser = argparse.ArgumentParser(
            description=self.description, usage=self.usage_str
        )

        self.argument_parser.add_argument(
            ".",
            metavar="config_file",
            nargs="?",
            help="Top-level config file to load (optional).",
            type=Path,
            default=MISSING,
        )

        self.argument_parser.add_argument(
            "--trial-index",
            type=self.match_trial_range,
            help="Execute only the specified trials.",
            default=MISSING,
            dest=self.global_config._trial_indices_key,
        )

        if not self.global_config.config_only:
            self.argument_parser.add_argument(
                "--cordage-comment",
                type=str,
                help="Add a comment to the annotation of this series.",
                default=MISSING,
                dest=self.global_config._experiment_comment_key,
                metavar="COMMENT",
            )

            self.argument_parser.add_argument(
                "--output_dir",
                type=Path,
                help="Path to use as the output directory.",
                default=MISSING,
                dest=self.global_config._output_dir_key,
                metavar="PATH",
            )

        self.arg_group_config = self.argument_parser.add_argument_group("configuration")
        self.add_arguments_to_parser(self.main_config_cls)

    def _add_argument_to_parser(self, arg_name: str, arg_type: Any, help: str, **kw):  # noqa: A002
        # If the field is also a dataclass, recurse (nested config)
        if dataclasses.is_dataclass(arg_type):
            assert isinstance(arg_type, type)
            self.arg_group_config.add_argument(
                f"--{arg_name}", type=Path, default=MISSING, help=help, metavar="PATH", **kw
            )
            self.add_arguments_to_parser(arg_type, prefix=arg_name)

        # Look the fields annotation to determine which type of argument
        # to add

        # Choice field
        elif get_origin(arg_type) is Literal:
            # Value must be from this set
            choices = get_args(arg_type)

            literal_arg_type = type(choices[0])

            if any(not isinstance(c, literal_arg_type) for c in choices):
                msg = f"If Literal is used, all values must be of the same type ({arg_name})."
                raise TypeError(msg)

            self.arg_group_config.add_argument(
                f"--{arg_name}",
                type=literal_arg_type,
                choices=choices,
                default=MISSING,
                help=help,
                **kw,
            )

        elif get_origin(arg_type) is Union:
            args = [arg for arg in get_args(arg_type) if arg is not type(None)]

            if len(args) == 1:
                # optional
                self._add_argument_to_parser(arg_name, args[0], help=help, **kw)

            else:
                msg = (
                    f"Parameter `{arg_name}` could not be processed:"
                    "Config parser does not support Union annotations with more than one type "
                    "other than None."
                )
                raise TypeError(msg)

        # Boolean field
        elif arg_type is bool:
            # Create a true/false flag -> the destination is identical
            self.arg_group_config.add_argument(
                f"--{arg_name}",
                action="store_true",
                default=MISSING,
                help=help + " (set the value to True)",
                **kw,
            )

            self.arg_group_config.add_argument(
                f"--not-{arg_name}",
                dest=arg_name,
                action="store_false",
                default=MISSING,
                help=help + " (set the value to False)",
                **kw,
            )

        elif arg_type in SUPPORTED_PRIMITIVES:
            self.arg_group_config.add_argument(
                f"--{arg_name}", type=arg_type, default=MISSING, help=help, **kw
            )

    def add_arguments_to_parser(self, config_cls: type, prefix: Optional[str] = None):
        """Add all fields in the (nested) config class to the parser.

        Recursively iterate over the fields adding arguments to the
        parser.
        """

        # read documentation of config dataclass. If no help metadata is
        # given, this will be used a the help text.
        param_doc = {}
        if config_cls.__doc__ is not None:
            # parse doc text, to generate help text for fields
            for param in parse_docstring(config_cls.__doc__).params:
                param_doc[param.arg_name] = param.description

        # Iterate over all fields in the dataclass to add arguments to
        # the parser
        for field in dataclasses.fields(config_cls):
            if not field.init:
                continue

            if prefix is None and field.name == self.global_config.param_name_output_dir:
                continue

            # Set prefixed argument name
            if prefix is not None:
                arg_name = f"{prefix}.{field.name}"
            else:
                arg_name = field.name

            # Retrieve help text
            help_text = field.metadata.get("help", param_doc.get(field.name, ""))

            self._add_argument_to_parser(arg_name, field.type, help=help_text)

    def remove_missing_values(self, data: Mapping) -> dict[str, Any]:
        return {k: v for k, v in data.items() if v is not MISSING}

    def construct_func_kwargs(self, trial: Trial):
        # construct arguments for the passed callable
        func_kw: dict[str, Any] = {}

        # check if any other parameters are expected which can be
        # resolved
        for name, param in self.func_parameters.items():
            assert param.kind != param.POSITIONAL_ONLY, (
                "Cordage currently does not support positional only parameters."
            )

            if name == self.global_config.param_name_config:
                # pass the configuration
                func_kw[name] = trial.config

            elif not self.global_config.config_only:
                if name == self.global_config.param_name_output_dir:
                    # pass path to output directory
                    if issubclass(param.annotation, str):
                        func_kw[name] = str(trial.output_dir)
                    else:
                        func_kw[name] = trial.output_dir

                elif name == self.global_config.param_name_trial_object:
                    # pass trial object
                    func_kw[name] = trial

        return func_kw

    def parse_args(self, args: Optional[list[str]] = None) -> Experiment:
        """Parse the command line arguments.

        Args:
            args: A list of arguments (usually from the CLI). If
              none, sys.argv[1:] is used.

        Returns:
            The resulting experiment (the resulting configuration of the
            experiment) is the result of the default values, a
            potentially loaded config file, and parameters passed via
            the CLI args.
        """

        if args is None:
            # args default to the system args
            args = sys.argv[1:]
        else:
            args = list(args)

        # construct parser
        argument_data: dict = vars(self.argument_parser.parse_args(args))
        argument_data = self.remove_missing_values(argument_data)

        conf_file_comment: Optional[str] = None
        cli_series_comment: Optional[str] = None

        series_kw = {
            "function": self.func_name,
            "global_config": self.global_config,
            "status": Status.PENDING,
            "additional_info": {"description": self.description, "parsed_arguments": args},
        }

        if not self.global_config.config_only:
            cli_series_comment = argument_data.pop(
                self.global_config._experiment_comment_key, None
            )
            series_kw["output_dir"] = argument_data.pop(self.global_config._output_dir_key, None)

        config_path = argument_data.pop(".", None)

        argument_data = nest_items(argument_data.items())

        if config_path is not None:
            new_conf_data = read_dict_from_file(config_path)

            new_conf_data = nest_items(new_conf_data.items())

            series_kw["series_spec"] = new_conf_data.pop(self.global_config._series_spec_key, None)

            nested_update(new_conf_data, argument_data)

            argument_data = new_conf_data

            if not self.global_config.config_only:
                # another series comment might be given via the confi
                # g file
                # in this case, the comments are added to another
                conf_file_comment = argument_data.pop(
                    self.global_config._experiment_comment_key, None
                )
        else:
            series_kw["series_spec"] = None

        # series skip might be given via the command line
        # ("--series-skip <n>") or a config file "__series-skip__"
        series_kw["trial_indices"] = argument_data.pop(self.global_config._trial_indices_key, None)
        series_kw["base_config"] = argument_data
        series_kw["config_cls"] = self.main_config_cls

        series: Series = Series(**series_kw)

        if not self.global_config.config_only:
            if cli_series_comment is not None and conf_file_comment is not None:
                series.comment = conf_file_comment + "\n\n" + cli_series_comment
            else:
                series.comment = conf_file_comment or cli_series_comment or ""

        logger.debug("%d experiments found in configuration", len(series))

        if series.is_singular:
            return next(iter(series))
        else:
            return series

    def from_configuration(
        self,
        config=None,
        base_config=None,
        series_spec=None,
        trial_indices: Optional[TrialIndices] = None,
        comment: Optional[str] = None,
    ) -> Experiment:
        _usage = "Either pass `config` or `base_config` and `series_spec`"

        if config is not None:
            assert base_config is None and series_spec is None and trial_indices is None, _usage

            trial: Trial = Trial(
                function=self.func_name,
                config=config,
                global_config=self.global_config,
                additional_info={"description": self.description},
            )

            trial.comment = comment

            return trial

        else:
            assert base_config is not None and series_spec is not None

            series: Series = Series(
                function=self.func_name,
                base_config=base_config,
                global_config=self.global_config,
                series_spec=series_spec,
                trial_indices=trial_indices,
                additional_info={"description": self.description},
            )
            series.comment = comment
            return series

    def execute(self, experiment: Experiment):
        """Execute a given experiment.

        The function of this `FunctionContext` is called with the
        configuration given by the experiment.
        """
        if isinstance(experiment, Trial):
            if self.global_config.config_only:
                # execute function with the constructed keyword
                # arguments
                func_kw = self.construct_func_kwargs(experiment)
                experiment.metadata.result = self.func(**func_kw)

            else:
                # only use stack if full feature-set is used
                with experiment_stack.with_experiment_on_stack(experiment):
                    func_kw = self.construct_func_kwargs(experiment)
                    experiment.metadata.result = self.func(**func_kw)

        elif isinstance(experiment, Series):
            if self.global_config.config_only:
                for trial in experiment:
                    self.execute(trial)
            else:
                with experiment_stack.with_experiment_on_stack(experiment):
                    for trial in experiment:
                        self.execute(trial)

        else:
            msg = "Passed object must be Trial or Series"
            raise TypeError(msg)
